@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker and Sean Charles
     E-mail:        jan@swi-prolog.org and <sean at objitsu dot com>
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2013-2020, Sean Charles
+    Copyright (c)  2013-2022, Sean Charles
                               SWI-Prolog Solutions b.v.
     All rights reserved.
 
@@ -74,20 +74,27 @@
             redis_current_command/3     % +Redis, +Command, -Properties
           ]).
 :- autoload(library(socket), [tcp_connect/3]).
-:- autoload(library(apply), [maplist/2, convlist/3, maplist/4, maplist/3]).
+:- autoload(library(apply), [maplist/2, convlist/3, maplist/3, maplist/5]).
 :- autoload(library(broadcast), [broadcast/1]).
 :- autoload(library(error),
             [ must_be/2,
+	      type_error/2,
               instantiation_error/1,
               uninstantiation_error/1,
-              existence_error/2
+              existence_error/2,
+              existence_error/3
             ]).
 :- autoload(library(lazy_lists), [lazy_list/2]).
 :- autoload(library(lists), [append/3, member/2]).
-:- autoload(library(option), [merge_options/3, option/2, option/3]).
+:- autoload(library(option), [merge_options/3, option/2,
+			      option/3, select_option/4]).
 :- autoload(library(pairs), [group_pairs_by_key/2]).
-:- use_module(library(debug), [debug/3]).
+:- autoload(library(time), [call_with_time_limit/2]).
+:- use_module(library(debug), [debug/3, assertion/1]).
 :- use_module(library(settings), [setting/4, setting/2]).
+:- if(exists_source(library(ssl))).
+:- autoload(library(ssl), [ssl_context/3, ssl_negotiate/5]).
+:- endif.
 
 :- use_foreign_library(foreign(redis4pl)).
 
@@ -95,6 +102,8 @@
            "Max number of retries").
 :- setting(max_retry_wait, number, 10,
            "Max time to wait between recovery attempts").
+:- setting(sentinel_timeout, number, 0.2,
+	   "Time to wait for a sentinel").
 
 :- predicate_options(redis_server/3, 3,
                      [ pass_to(redis:redis_connect/3, 3)
@@ -139,7 +148,8 @@ User = "Bob"
 
 :- dynamic server/3.
 
-:- dynamic ( connection/2               % ServerName, Stream
+:- dynamic ( connection/2,              % ServerName, Stream
+	     sentinel/2			% Pool, Address
            ) as volatile.
 
 %!  redis_server(+ServerName, +Address, +Options) is det.
@@ -182,6 +192,24 @@ server(default, localhost:6379, []).
 %       Specify the connection protocol version.  Initially this is
 %       version 2.  Redis 6 also supports version 3.  When specified
 %       as `3`, the `HELLO` command is used to upgrade the protocol.
+%     - tls(true)
+%       When specified, initiate a TLS connection.  If this option is
+%       specified we must also specify the `cacert`, `key` and `cert`
+%       options.
+%     - cacert(+File)
+%       CA Certificate file to verify with.
+%     - cert(+File)
+%       Client certificate to authenticate with.
+%     - key(+File)
+%       Private key file to authenticate with.
+%     - sentinels(+ListOfAddresses)
+%       Used together with an Address of the form sentinel(MasterName)
+%       to enable contacting a network of Redis servers guarded by a
+%       sentinel network.
+%     - sentinel_user(+User)
+%     - sentinel_password(+Password)
+%       Authentication information for the senitels.  When omitted we
+%       try to connect withour authentication.
 %
 %   Instead of using these predicates, redis/2  and redis/3 are normally
 %   used with a _server name_  argument registered using redis_server/3.
@@ -222,9 +250,12 @@ redis_connect(Address, Conn, Options) :-
 %
 %       redis_connection(Id, Stream, Failures, Options)
 
-do_connect(Id, Address0, Conn, Options) :-
+do_connect(Id, sentinel(Pool), Conn, Options) =>
+    sentinel_master(Id, Pool, Conn, Options).
+do_connect(Id, Address0, Conn, Options) =>
     tcp_address(Address0, Address),
-    tcp_connect(Address, Stream, Options),
+    tcp_connect(Address, Stream0, Options),
+    tls_upgrade(Address, Stream0, Stream, Options),
     Conn = redis_connection(Id, Stream, 0, Options),
     hello(Conn, Options).
 
@@ -232,6 +263,113 @@ tcp_address(unix(Path), Path) :-
     !.                                  % Using an atom is ambiguous
 tcp_address(Address, Address).
 
+%!  tls_upgrade(+Address, +Raw, -Stream, +Options) is det.
+%
+%   Upgrade to a TLS connection when tls(true) is specified.
+
+:- if(current_predicate(ssl_context/3)).
+tls_upgrade(Host:_Port, Raw, Stream, Options) :-
+    option(tls(true), Options),
+    !,
+    must_have_option(cacert(CacertFile), Options),
+    must_have_option(key(KeyFile), Options),
+    must_have_option(cert(CertFile), Options),
+    ssl_context(client, SSL,
+		[ host(Host),
+		  certificate_file(CertFile),
+		  key_file(KeyFile),
+		  cacerts([file(CacertFile)]),
+		  cert_verify_hook(tls_verify),
+		  close_parent(true)
+		]),
+    stream_pair(Raw, RawRead, RawWrite),
+    ssl_negotiate(SSL, RawRead, RawWrite, Read, Write),
+    stream_pair(Stream, Read, Write).
+:- endif.
+tls_upgrade(_, Stream, Stream, _).
+
+:- if(current_predicate(ssl_context/3)).
+
+%!  tls_verify(+SSL, +ProblemCert, +AllCerts, +FirstCert, +Status) is semidet.
+%
+%   Accept  or reject  the certificate  verification.  Similar  to the
+%   Redis  command   line  client   (``redis-cli``),  we   accept  the
+%   certificate as long as it is signed, not verifying the hostname.
+
+:- public tls_verify/5.
+tls_verify(_SSL, _ProblemCert, _AllCerts, _FirstCert, verified) :-
+    !.
+tls_verify(_SSL, _ProblemCert, _AllCerts, _FirstCert, hostname_mismatch) :-
+    !.
+tls_verify(_SSL, _ProblemCert, _AllCerts, _FirstCert, _Error) :-
+    fail.
+
+:- endif.
+
+%!  sentinel_master(+ServerId, +SetinelPool, -Connection, +Options) is det.
+%
+%   Discover the master and connect to it.
+
+sentinel_master(Id, Pool, Master, Options) :-
+    must_have_option(sentinels(Sentinels), Options),
+    sentinel_auth(Options, Options1),
+    setting(sentinel_timeout, TMO),
+    (   sentinel(Pool, Sentinel)
+    ;   member(Sentinel, Sentinels)
+    ),
+    catch(call_with_time_limit(
+	      TMO,
+	      do_connect(Id, Sentinel, Conn,
+			 [sentinel(true)|Options1])),
+	  Error,
+	  (print_message(warning, Error),fail)),
+    !,
+    debug(redis(sentinel), 'Connected to sentinel at ~p', [Sentinel]),
+    call_cleanup(
+	query_sentinel(Pool, Conn, Sentinel, MasterAddr),
+	redis_disconnect(Conn)),
+    debug(redis(sentinel), 'Sentinel claims master is at ~p', [MasterAddr]),
+    do_connect(Id, MasterAddr, Master, Options),
+    debug(redis(sentinel), 'Connected to claimed master', []),
+    redis(Master, role, Role),
+    (   Role = [master|_Slaves]
+    ->  debug(redis(sentinel), 'Verified role at ~p', [MasterAddr])
+    ;   redis_disconnect(Master),
+	debug(redis(sentinel), '~p is not the master: ~p', [MasterAddr, Role]),
+	sleep(TMO),
+	sentinel_master(Id, Pool, Master, Options)
+    ).
+
+sentinel_auth(Options0, Options) :-
+    option(sentinel_user(User), Options0),
+    option(sentinel_password(Passwd), Options0),
+    !,
+    merge_options([user(User), password(Passwd)], Options0, Options).
+sentinel_auth(Options0, Options) :-
+    select_option(password(_), Options0, Options, _).
+
+
+query_sentinel(Pool, Conn, Sentinel, Host:Port) :-
+    redis(Conn, sentinel('get-master-addr-by-name', Pool), MasterData),
+    MasterData = [Host,Port],
+    redis(Conn, sentinel(sentinels, Pool), Peers),
+    transaction(update_known_sentinels(Pool, Sentinel, Peers)).
+
+update_known_sentinels(Pool, Sentinel, Peers) :-
+    retractall(sentinel(Pool, _)),
+    maplist(update_peer_sentinel(Pool), Peers),
+    asserta(sentinel(Pool, Sentinel)).
+
+update_peer_sentinel(Pool, Attrs),
+  memberchk(ip-Host, Attrs),
+  memberchk(port-Port, Attrs) =>
+    asserta(sentinel(Pool, Host:Port)).
+
+must_have_option(Opt, Options) :-
+    option(Opt, Options),
+    !.
+must_have_option(Opt, Options) :-
+    existence_error(option, Opt, Options).
 
 %!  hello(+Connection, +Option)
 %
@@ -399,16 +537,20 @@ redis(Redis, Req) :-
 %       strings or integers (__no floats__) is translated into
 %       a string `"A:B:..."`.  This is a common shorthand for
 %       representing Redis keys.
-%     - A term prolog(Term) is emitted as "\u0000T\u0000" followed
+%     - A term Term as prolog is emitted as "\u0000T\u0000" followed
 %       by Term in canonical form.
 %     - Any other term is emitted as write/1.
 %
 %   Reply is either a plain term (often a  variable) or a term `Value as
 %   Type`. In the latter form,  `Type`   dictates  how  the Redis _bulk_
-%   reply is translated to Prolog. The default equals to string(utf8).
+%   reply is translated to Prolog. The default equals to `auto`, i.e.,
+%   as a number of the content satisfies the Prolog number syntax and
+%   as an atom otherwise.
 %
-%     - status(String)
-%       Returned if the server replies with ``+ Status``.
+%     - status(Atom)
+%       Returned if the server replies with ``+ Status``.  Atom
+%       is the textual value of `Status` converted to lower case,
+%       e.g., status(ok) or status(pong).
 %     - `nil`
 %       This atom is returned for a NIL/NULL value.  Note that if
 %       the reply is only `nil`, redis/3 _fails_.  The `nil` value
@@ -569,8 +711,10 @@ read_pipeline(Redis, S, PipeLine) :-
     catch(read_pipeline2(Redis, S, PipeLine), E, true),
     (   var(Formal)
     ->  true
-    ;   print_message(warning, E),
-        redis_disconnect(Redis, [force(true)]),
+    ;   reconnect_error(E)
+    ->  redis_disconnect(Redis, [force(true)]),
+        throw(E)
+    ;   resync(Redis),
         throw(E)
     ).
 
@@ -608,13 +752,14 @@ bind_reply(_Command, _).
 :- meta_predicate recover(+, +, 0).
 
 recover(Error, Redis, Goal) :-
-    reconnect_error(Error),
+    Error = error(Formal, _),
+    reconnect_error(Formal),
     auto_reconnect(Redis),
     !,
     debug(redis(recover), '~p: got error ~p; trying to reconnect',
           [Redis, Error]),
     redis_disconnect(Redis, [force(true)]),
-    (   wait(Redis)
+    (   wait_to_retry(Redis, Error)
     ->  call(Goal),
         retractall(failure(Redis, _))
     ;   throw(Error)
@@ -630,10 +775,12 @@ auto_reconnect(Server) :-
     server(Server, _, Options),
     option(reconnect(true), Options, true).
 
-reconnect_error(error(socket_error(_Code, _),_)).
-reconnect_error(error(syntax_error(unexpected_eof),_)).
+reconnect_error(io_error(_Action, _On)).
+reconnect_error(socket_error(_Code, _)).
+reconnect_error(syntax_error(unexpected_eof)).
+reconnect_error(existence_error(stream, _)).
 
-%!  wait(+Redis)
+%!  wait(+Redis, +Error)
 %
 %   Wait for some time after a failure. First  we wait for 10ms. This is
 %   doubled on each failure upto the   setting  `max_retry_wait`. If the
@@ -642,7 +789,7 @@ reconnect_error(error(syntax_error(unexpected_eof),_)).
 
 :- dynamic failure/2 as volatile.
 
-wait(Redis) :-
+wait_to_retry(Redis, Error) :-
     redis_failures(Redis, Failures),
     setting(max_retry_count, Count),
     Failures < Count,
@@ -651,6 +798,8 @@ wait(Redis) :-
     setting(max_retry_wait, MaxWait),
     Wait is min(MaxWait*100, 1<<Failures)/100.0,
     debug(redis(recover), '  Sleeping ~p seconds', [Wait]),
+    retry_message_level(Failures, Level),
+    print_message(Level, redis(retry(Redis, Failures, Wait, Error))),
     sleep(Wait).
 
 redis_failures(redis_connection(_,_,Failures0,_), Failures) :-
@@ -672,6 +821,9 @@ redis_set_failures(Server, Count) :-
     retractall(failure(Server, _)),
     asserta(failure(Server, Count)).
 
+retry_message_level(0, warning) :- !.
+retry_message_level(_, silent).
+
 
 %!  redis(+Request)
 %
@@ -715,7 +867,7 @@ redis_read(Redis, Reply) :-
 %   useful for relatively short lists holding possibly large items.
 %
 %   Note that values retrieved are _strings_, unless the value was added
-%   using prolog(Term).
+%   using `Term as prolog`.
 %
 %   @see lazy_list/2 for a discussion  on   the  difference between lazy
 %   lists and normal lists.
@@ -783,11 +935,10 @@ redis_set_list(Redis, Key, List) :-
 %   by key. redis_get_hash/3 is notably a   user friendly alternative to
 %   the Redis ``HGETALL`` command. If the  Redis   hash  is  not used by
 %   other (non-Prolog) applications one  may   also  consider  using the
-%   prolog(Term) syntax to store the Prolog dict as-is.
+%   `Term as prolog` syntax to store the Prolog dict as-is.
 
 redis_get_hash(Redis, Key, Dict) :-
-    redis(Redis, hgetall(Key), TwoList),
-    redis_array_dict(TwoList, _, Dict).
+    redis(Redis, hgetall(Key), Dict as dict(auto)).
 
 redis_set_hash(Redis, Key, Dict) :-
     redis_array_dict(Array, _, Dict),
@@ -884,11 +1035,10 @@ scan_next(State, List, Tail) :-
     Term =.. CList2,
     redis(Redis, Term, [NewCursor,Elems0]),
     scan_pairs(Command, Elems0, Elems),
-    (   NewCursor == "0"
+    (   NewCursor == 0
     ->  List = Elems,
         Tail = []
-    ;   number_string(CursorI, NewCursor),
-        nb_setarg(3, State, CursorI),
+    ;   nb_setarg(3, State, NewCursor),
         append(Elems, Tail, List)
     ).
 
@@ -1057,6 +1207,10 @@ pubsub_clean(Id) :-
 %
 %   Add/remove channels from for the   subscription. If no subscriptions
 %   remain, the listening thread terminates.
+%
+%   @arg Channels is either a single  channel   or  a list thereof. Each
+%   channel specification is either an atom   or a term `A:B:...`, where
+%   all parts are atoms.
 
 redis_subscribe(Id, Spec) :-
     channels(Spec, Channels),
@@ -1090,12 +1244,35 @@ redis_current_subscription(Id, Channels) :-
     group_pairs_by_key(Sorted, Grouped),
     member(Id-Channels, Grouped).
 
-channels(List, List) :-
-    is_list(List),
+channels(Spec, List) :-
+    is_list(Spec),
     !,
-    must_be(list(atom), List).
-channels(Ch, [Ch]) :-
-    must_be(atom, Ch).
+    maplist(channel_name, Spec, List).
+channels(Ch, [Key]) :-
+    channel_name(Ch, Key).
+
+channel_name(Atom, Atom) :-
+    atom(Atom),
+    !.
+channel_name(Key, Atom) :-
+    phrase(key_parts(Key), Parts),
+    !,
+    atomic_list_concat(Parts, :, Atom).
+channel_name(Key, _) :-
+    type_error(redis_key, Key).
+
+key_parts(Var) -->
+    { var(Var), !, fail }.
+key_parts(Atom) -->
+    { atom(Atom) },
+    !,
+    [Atom].
+key_parts(A:B) -->
+    key_parts(A),
+    key_parts(B).
+
+
+
 
 register_subscription(Id, Channel) :-
     (   subscription(Id, Channel)
@@ -1123,15 +1300,17 @@ redis_listen_loop(Redis, Id, Conn) :-
     ;   true
     ).
 
-redis_broadcast(_, ["subscribe", _Channel, _N]) :-
+redis_broadcast(_, [subscribe, _Channel, _N]) :-
     !.
-redis_broadcast(Redis, ["message", ChannelS, Data]) :-
+redis_broadcast(Redis, [message, Channel, Data]) :-
     !,
-    atom_string(Channel, ChannelS),
     catch(broadcast(redis(Redis, Channel, Data)),
           Error,
           print_message(error, Error)).
 redis_broadcast(Redis, Message) :-
+    assertion((Message = [Type, Channel, _Data],
+               atom(Type),
+               atom(Channel))),
     debug(redis(warning), '~p: Unknown message while listening: ~p',
           [Redis,Message]).
 
@@ -1162,10 +1341,10 @@ redis_read_stream(Redis, SI, Out) :-
     ->  handle_push_messages(Push, Redis),
         (   var(Error)
         ->  Out = Out0
-        ;   throw(Error)
+        ;   resync(Redis),
+            throw(Error)
         )
-    ;   print_message(warning, E),
-        redis_disconnect(Redis, [force(true)]),
+    ;   redis_disconnect(Redis, [force(true)]),
         throw(E)
     ).
 
@@ -1180,6 +1359,35 @@ handle_push_messages([H|T], Redis) :-
 
 handle_push_message(["pubsub"|List], Redis) :-
     redis_broadcast(Redis, List).
+% some protocol version 3 push messages (such as
+% __keyspace@* events) seem to come directly
+% without a pubsub header
+handle_push_message([message|List], Redis) :-
+    redis_broadcast(Redis, [message|List]).
+
+
+%!  resync(+Redis) is det.
+%
+%   Re-synchronize  after  an  error.  This  may  happen  if  some  type
+%   conversion fails and we have read  a   partial  reply. It is hard to
+%   figure out what to read from where we are, so we echo a random magic
+%   sequence and read until we find the reply.
+
+resync(Redis) :-
+    E = error(Formal,_),
+    catch(do_resync(Redis), E, true),
+    (   var(Formal)
+    ->  true
+    ;   redis_disconnect(Redis, [force(true)])
+    ).
+
+do_resync(Redis) :-
+    A is random(1_000_000_000),
+    redis_stream(Redis, S, true),
+    redis_write_msg(S, echo(A)),
+    catch(call_with_time_limit(0.2, '$redis_resync'(S, A)),
+          time_limit_exceeded,
+          throw(error(time_limit_exceeded,_))).
 
 
 %!  redis_read_msg(+Stream, -Message, -Error, -PushMessages) is det.
@@ -1199,7 +1407,12 @@ handle_push_message(["pubsub"|List], Redis) :-
 		 *******************************/
 
 :- multifile
-    prolog:error_message//1.
+    prolog:error_message//1,
+    prolog:message//1.
 
 prolog:error_message(redis_error(Code, String)) -->
     [ 'REDIS: ~w: ~s'-[Code, String] ].
+
+prolog:message(redis(retry(_Redis, _Failures, Wait, Error))) -->
+    [ 'REDIS: connection error.  Retrying in ~2f seconds'-[Wait], nl ],
+    [ '    '-[] ], '$messages':translate_message(Error).
