@@ -49,8 +49,8 @@
 
             py_func/3,                  % +Module, +Func, -Return
             py_func/4,                  % +Module, +Func, -Return, +Options
-            py_dot/4,                   % +Module, +ObjRef, +Meth, ?Ret
-            py_dot/5,                   % +Module, +ObjRef, +Meth, -Ret, +Options
+            py_dot/3,                   % +ObjRef, +Meth, ?Ret
+            py_dot/4,                   % +ObjRef, +Meth, -Ret, +Options
 
             values/3,                   % +Dict, +Path, ?Val
             keys/2,                     % +Dict, ?Keys
@@ -63,8 +63,14 @@
             py_pp/2,                    % +Stream, +Term
             py_pp/3,                    % +Stream, +Term, +Options
 
-            py_obj_dir/2,               % +ObjRef,-List
-            py_obj_dict/2,              % obj_dict(+ObjRef, -Dict)
+            py_object_dir/2,            % +ObjRef, -List
+            py_object_dict/2,           % +ObjRef, -Dict
+            py_obj_dir/2,               % +ObjRef, -List (deprecated)
+            py_obj_dict/2,              % +ObjRef, -Dict (deprecated)
+            py_type/2,			% +ObjRef, -Type:atom
+            py_isinstance/2,            % +ObjRef, +Type
+            py_module_exists/1,         % +Module
+            py_hasattr/2,               % +Module, ?Symbol
 
             py_module/2,                % +Module:atom, +Source:string
 
@@ -110,12 +116,10 @@ add_python_dll_dir :-
                        py_string_as(oneof([string,atom]))
                      ]).
 :- predicate_options(py_func/4, 4,
-                     [ sizecheck(boolean),
-                       pass_to(py_call/3, 3)
+                     [ pass_to(py_call/3, 3)
                      ]).
 :- predicate_options(py_dot/5, 5,
-                     [ sizecheck(boolean),
-                       pass_to(py_call/3, 3)
+                     [ pass_to(py_call/3, 3)
                      ]).
 
 :- public
@@ -126,6 +130,7 @@ add_python_dll_dir :-
 
 :- create_prolog_flag(py_backtrace,       true, [type(boolean), keep(true)]).
 :- create_prolog_flag(py_backtrace_depth, 4,    [type(integer), keep(true)]).
+:- create_prolog_flag(py_argv,		  [],   [type(term), keep(true)]).
 
 /** <module> Call Python from Prolog
 
@@ -348,7 +353,7 @@ py_version :-
 %   True if Term is a Prolog term that represents a Python dict.
 %
 %   @compat PIP. The SWI-Prolog version accepts   both a SWI-Prolog dict
-%   as the \{k:v,\ldots\} representation.  See   `py_dict_as`  option of
+%   and the `{k:v,...}`  representation.  See   `py_dict_as`  option  of
 %   py_call/2.
 
 py_is_dict(Dict), is_dict(Dict) => true.
@@ -422,19 +427,18 @@ py_func(Module, Function, Return) :-
 py_func(Module, Function, Return, Options) :-
     py_call(Module:Function, Return, Options).
 
-%!  py_dot(+Module, +ObjRef, +MethAttr, -Ret) is det.
-%!  py_dot(+Module, +ObjRef, +MethAttr, -Ret, +Options) is det.
+%!  py_dot(+ObjRef, +MethAttr, -Ret) is det.
+%!  py_dot(+ObjRef, +MethAttr, -Ret, +Options) is det.
 %
 %   Call a method or access  an  attribute   on  the  object ObjRef. The
 %   SWI-Prolog implementation is equivalent  to py_call(ObjRef:MethAttr,
 %   Return). See py_call/2 for details.
 %
-%   @arg Module is ignored (why do we need that if we have ObjRef?)
 %   @compat PIP.  See py_func/3 for details.
 
-py_dot(_Module, ObjRef, MethAttr, Ret) :-
+py_dot(ObjRef, MethAttr, Ret) :-
     py_call(ObjRef:MethAttr, Ret).
-py_dot(_Module, ObjRef, MethAttr, Ret, Options) :-
+py_dot(ObjRef, MethAttr, Ret, Options) :-
     py_call(ObjRef:MethAttr, Ret, Options).
 
 
@@ -633,21 +637,99 @@ opts_kws(Options, Kws) :-
 pair_kws(Name-Value, Name=Value).
 
 
-%!  py_obj_dir(+ObjRef, -List) is det.
-%!  py_obj_dict(+ObjRef, -Dict) is det.
+%!  py_object_dir(+ObjRef, -List) is det.
+%!  py_object_dict(+ObjRef, -Dict) is det.
 %
-%   Examine attributes of an object.  The predicate py_obj_dir/2 fetches
-%   the names of all attributes, while   py_obj_dict/2  gets a dict with
-%   all attributes and their values.
+%   Examine attributes of  an  object.   The  predicate  py_object_dir/2
+%   fetches the names of all attributes,   while  py_object_dir/2 gets a
+%   dict with all attributes and their values.
 %
 %   @compat PIP
 
-py_obj_dir(ObjRef, List) :-
+py_object_dir(ObjRef, List) :-
     py_call(ObjRef:'__dir__'(), List).
 
-py_obj_dict(ObjRef, Dict) :-
+py_object_dict(ObjRef, Dict) :-
     py_call(ObjRef:'__dict__', Dict).
 
+%!  py_obj_dir(+ObjRef, -List) is det.
+%!  py_obj_dict(+ObjRef, -Dict) is det.
+%
+%   @deprecated Use py_object_dir/2 or py_object_dict/2.
+
+py_obj_dir(ObjRef, List) :-
+    py_object_dir(ObjRef, List).
+
+py_obj_dict(ObjRef, Dict) :-
+    py_object_dict(ObjRef, Dict).
+
+
+%!  py_type(+ObjRef, -Type:atom) is det.
+%
+%   True when Type is the name of the   type of ObjRef. This is the same
+%   as ``type(ObjRef).__name__`` in Python.
+%
+%   @compat PIP
+
+py_type(ObjRef, Type) :-
+    py_call(type(ObjRef):'__name__', Type).
+
+%!  py_isinstance(+ObjRef, +Type) is semidet.
+%
+%   True if ObjRef is an instance of Type   or an instance of one of the
+%   sub types of Type. This  is   the  same as ``isinstance(ObjRef)`` in
+%   Python.
+%
+%   @arg Type is either a term `Module:Type` or a plain atom to refer to
+%   a built-in type.
+%
+%   @compat PIP
+
+py_isinstance(Obj, Module:Type) =>
+    py_call(isinstance(Obj, eval(Module:Type)), @true).
+py_isinstance(Obj, Type) =>
+    py_call(isinstance(Obj, eval(sys:modules:'__getitem__'(builtins):Type)), @true).
+
+%!  py_module_exists(+Module) is semidet.
+%
+%   True if Module is a currently  loaded   Python  module  or it can be
+%   loaded.
+%
+%   @compat PIP
+
+py_module_exists(Module) :-
+    must_be(atom, Module),
+    py_call(sys:modules:'__contains__'(Module), @true),
+    !.
+py_module_exists(Module) :-
+    py_call(importlib:util:find_spec(Module), R),
+    R \== @none,
+    py_free(R).
+
+%!  py_hasattr(+ModuleOrObj, ?Name) is nondet.
+%
+%   True when Name is an attribute of   Module. The name is derived from
+%   the Python built-in hasattr(). If Name   is unbound, this enumerates
+%   the members of py_object_dir/2.
+%
+%   @arg ModuleOrObj If this is an atom it refers to a module, otherwise
+%   it must be a Python object reference.
+%
+%   @compat PIP
+
+py_hasattr(ModuleOrObj, Name) :-
+    var(Name),
+    !,
+    py_object_dir(ModuleOrObj, Names),
+    member(Name, Names).
+py_hasattr(ModuleOrObj, Name) :-
+    must_be(atom, Name),
+    (   atom(ModuleOrObj)
+    ->  py_call(ModuleOrObj:'__name__'), % force loading
+        py_call(hasattr(eval(sys:modules:'__getitem__'(ModuleOrObj)), Name), @true)
+    ;   py_call(hasattr(ModuleOrObj, Name), @true)
+    ).
+
 
 %!  py_module(+Module:atom, +Source:string) is det.
 %
@@ -703,12 +785,13 @@ py_initialize :-
     access_file(Cfg, read),
     !,
     current_prolog_flag(executable, Program),
-    py_initialize(Program, ['-I'], []),
+    current_prolog_flag(py_argv, Argv),
+    py_initialize(Program, ['-I'|Argv], []),
     py_call(sys:prefix = VEnv),
     venv_update_path(VEnvDir).
 py_initialize :-
     current_prolog_flag(executable, Program),
-    current_prolog_flag(argv, Argv),
+    current_prolog_flag(py_argv, Argv),
     py_initialize(Program, Argv, []).
 
 venv_update_path(VEnvDir) :-
@@ -744,8 +827,8 @@ no_env_dir('dist-packages').
 %   Initialize  and configure  the  embedded Python  system.  If  this
 %   predicate is  not called before any  other call to Python  such as
 %   py_call/2, it is called _lazily_, passing the Prolog executable as
-%   Program, the  non-Prolog arguments  as Argv  and an  empty Options
-%   list.
+%   Program, passing Argv from the  Prolog flag `py_argv` and an empty
+%   Options list.
 %
 %   Calling this predicate while the  Python is already initialized is
 %   a  no-op.  This  predicate is  thread-safe, where  the first  call
