@@ -131,7 +131,7 @@ swipl_call(PyObject *self, PyObject *args)
     return NULL;
   }
 
-  if ( PL_thread_attach_engine(NULL) < 0 )
+  if ( PL_thread_attach_engine(NULL) == -1 )
   { Py_SetPrologErrorFromChars("Cannot create thread");
     return NULL;
   }
@@ -398,7 +398,7 @@ swipl_open_query(PyObject *self, PyObject *args)
     return NULL;
   }
 
-  if ( PL_thread_attach_engine(NULL) < 0 )
+  if ( PL_thread_attach_engine(NULL) == -1 )
   { Py_SetPrologErrorFromChars("Cannot create thread");
     return NULL;
   }
@@ -696,5 +696,6 @@ static struct PyModuleDef swipl_module =
 
 PyMODINIT_FUNC
 PyInit__swipl(void)
-{ return PyModule_Create(&swipl_module);
+{ py_module_initialize_done = TRUE;
+  return PyModule_Create(&swipl_module);
 }
