@@ -1,3 +1,5 @@
+#!/bin/sh
+
 app=$1
 
 if [ -z "$app" ]; then
@@ -18,7 +20,7 @@ fi
 
 printf "Adding Macport dylibs to modules in $moduledir\n"
 
-changeset="$(echo $moduledir/*)"
+changeset="$(echo $moduledir/*) $app/Contents/MacOS/swipl"
 while [ ! -z "$changeset" ]; do
   newchanges=
   for f in $changeset; do
