@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2019, University of Amsterdam
+    Copyright (c)  1985-2023, University of Amsterdam
                               VU University Amsterdam
 			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -45,6 +46,7 @@
 #include "pl-gc.h"
 #include "pl-util.h"
 #include "pl-pro.h"
+#include "pl-modul.h"
 
 #undef LD
 #define LD LOCAL_LD
@@ -53,15 +55,27 @@
 
 #define PROFTYPE_MAGIC 0x639a2fb1
 
+#if USE_LD_MACROS
+#define	profile(count)		LDFUNC(profile, count)
+#define	thread_prof_ticks(_)	LDFUNC(thread_prof_ticks, _)
+#endif
+
+#define LDFUNC_DECLARATIONS
+
 static int  identify_def(term_t t, void *handle);
 static int  get_def(term_t t, void **handle);
-static void profile(intptr_t count, PL_local_data_t *__PL_ld);
+static void prof_release_def(void *handle);
+static void profile(intptr_t count);
+static int  thread_prof_ticks(void);
+
+#undef LDFUNC_DECLARATIONS
 
 static PL_prof_type_t prof_default_type =
 { identify_def,					/* unify a Definition */
-  get_def,
-  NULL,						/* dummy */
-  PROFTYPE_MAGIC
+  get_def,					/* Find a definition */
+  NULL,						/* Acquire */
+  prof_release_def,				/* Release a definition */
+  .magic = PROFTYPE_MAGIC
 };
 
 #define MAX_PROF_TYPES 10
@@ -70,25 +84,36 @@ static PL_prof_type_t *types[MAX_PROF_TYPES] = { &prof_default_type };
 #define PROFNODE_MAGIC 0x7ae38f24
 
 typedef struct call_node
-{ intptr_t	    magic;		/* PROFNODE_MAGIC */
+{ int		    magic;		/* PROFNODE_MAGIC */
   struct call_node *parent;
   void *            handle;		/* handle to procedure-id */
   PL_prof_type_t   *type;
-  uintptr_t	    calls;		/* Calls from the parent */
-  uintptr_t	    redos;		/* redos while here */
-  uintptr_t	    exits;		/* exits to the parent */
-  uintptr_t	    recur;		/* recursive calls */
-  uintptr_t	    ticks;		/* time-statistics */
-  uintptr_t	    sibling_ticks;	/* ticks in a siblings */
+  uint64_t	    calls;		/* Calls from the parent */
+  uint64_t	    fails;		/* fails back to choicepoint */
+  uint64_t	    exits;		/* exits to the parent */
+  uint64_t	    recur;		/* recursive calls */
+  uint64_t	    ticks;		/* time-statistics */
+  uint64_t	    sibling_ticks;	/* ticks in a siblings */
   struct call_node *next;		/* next in siblings chain */
   struct call_node *siblings;		/* my offspring */
 } call_node;
 
+#if USE_LD_MACROS
+#define	collectSiblingsTime(_)	LDFUNC(collectSiblingsTime, _)
+#endif /*USE_LD_MACROS*/
+
+#define LDFUNC_DECLARATIONS
+
 static void	freeProfileData(void);
-static void	collectSiblingsTime(ARG1_LD);
+static void	collectSiblingsTime(void);
+
+#undef LDFUNC_DECLARATIONS
+
+#define WITH_LD_IF_PROFILING(_)		WITH_LD(GD->profile.thread) if(LD _)
+#define WITH_LD_IF_PROFILING_AND(cond)	WITH_LD_IF_PROFILING( && (cond))
 
 int
-activateProfiler(prof_status active ARG_LD)
+activateProfiler(DECL_LD prof_status active)
 { int i;
   PL_local_data_t *profiling;
 
@@ -117,7 +142,7 @@ activateProfiler(prof_status active ARG_LD)
   if ( active )
   { LD->profile.time_at_last_tick =
     LD->profile.time_at_start     = active == PROF_CPU
-					? ThreadCPUTime(LD, CPU_USER)
+					? ThreadCPUTime(CPU_USER)
 					: WallTime();
 
     GD->profile.thread = LD;
@@ -136,12 +161,12 @@ activateProfiler(prof_status active ARG_LD)
 
 
 static int
-thread_prof_ticks(PL_local_data_t *ld)
-{ double t0 = ld->profile.time_at_last_tick;
-  double t1 = ld->profile.active == PROF_CPU ? ThreadCPUTime(ld, CPU_USER)
-				             : WallTime();
+thread_prof_ticks(DECL_LD)
+{ double t0 = LD->profile.time_at_last_tick;
+  double t1 = LD->profile.active == PROF_CPU ? ThreadCPUTime(CPU_USER)
+					     : WallTime();
 
-  ld->profile.time_at_last_tick = t1;
+  LD->profile.time_at_last_tick = t1;
 
   DEBUG(MSG_PROF_TICKS,
 	Sdprintf("%d ms\n", (int)((t1-t0)*1000.0)));
@@ -152,6 +177,8 @@ thread_prof_ticks(PL_local_data_t *ld)
 
 #ifdef __WINDOWS__
 
+#include <math.h>
+
 static UINT         timer;			/* our MM timer */
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -164,16 +191,14 @@ typedef DWORD DWORD_PTR;
 
 static void CALLBACK
 callTimer(UINT id, UINT msg, DWORD_PTR dwuser, DWORD_PTR dw1, DWORD_PTR dw2)
-{ PL_local_data_t *ld;
-
-  if ( (ld=GD->profile.thread) )
+{ WITH_LD_IF_PROFILING()
   { int newticks;
 
-    if ( (newticks = thread_prof_ticks(ld)) )
+    if ( (newticks = thread_prof_ticks()) )
     { if ( newticks < 0 )			/* Windows 95/98/... */
 	newticks = 1;
     }
-    profile(newticks, ld);
+    profile(newticks);
   }
 }
 
@@ -185,7 +210,7 @@ startProfiler(prof_status how)
   (void)how;
 
   rval = timeSetEvent(10,
-		      5,		/* resolution (milliseconds) */
+		      round(LD->profile.sample_period/1000),		/* resolution (milliseconds) */
 		      callTimer,
 		      (DWORD_PTR)0,
 		      TIME_PERIODIC);
@@ -195,7 +220,7 @@ startProfiler(prof_status how)
   else
     return PL_error(NULL, 0, NULL, ERR_SYSCALL, "timeSetEvent");
 
-  return activateProfiler(PROF_CPU PASS_LD);
+  return activateProfiler(PROF_CPU);
 }
 
 
@@ -230,21 +255,20 @@ static int timer_signal;		/* SIG* */
 
 static void
 sig_profile(int sig)
-{ PL_local_data_t *ld;
-  (void)sig;
+{ (void)sig;
 
 #if !defined(BSD_SIGNALS) && !defined(HAVE_SIGACTION)
   signal(SIGPROF, sig_profile);
 #endif
 
-  if ( (ld=GD->profile.thread) )
+  WITH_LD_IF_PROFILING()
   { int newticks;
 
-    if ( (newticks = thread_prof_ticks(ld)) )
+    if ( (newticks = thread_prof_ticks()) )
     { if ( newticks < 0 )			/* Windows 95/98/... */
 	newticks = 1;
     }
-    profile(newticks, ld);
+    profile(newticks);
   }
 }
 
@@ -266,16 +290,16 @@ startProfiler(prof_status how)
   timer_signal = sig;
 
   value.it_interval.tv_sec  = 0;
-  value.it_interval.tv_usec = 5000;		/* 5ms for real; also ok for cpu */
+  value.it_interval.tv_usec = LD->profile.sample_period;  // period in usecs.
   value.it_value.tv_sec  = 0;			/* on systems where 0 means now */
-  value.it_value.tv_usec = 5000;
+  value.it_value.tv_usec = LD->profile.sample_period;
 
   if ( setitimer(timer, &value, &ovalue) != 0 )
     return PL_error(NULL, 0, MSG_ERRNO, ERR_SYSCALL, setitimer);
 
   itimer = timer;
 
-  return activateProfiler(how PASS_LD);
+  return activateProfiler(how);
 }
 
 void
@@ -299,17 +323,14 @@ stopItimer(void)
 
 static int
 stopProfiler(void)
-{ PL_local_data_t *ld;
-
-  if ( (ld=GD->profile.thread) &&
-       ld->profile.active )
-  { double tend = ld->profile.active == PROF_CPU ? ThreadCPUTime(ld, CPU_USER)
+{ WITH_LD_IF_PROFILING_AND(LD->profile.active)
+  { double tend = LD->profile.active == PROF_CPU ? ThreadCPUTime(CPU_USER)
 						 : WallTime();
 
-    ld->profile.time += tend - ld->profile.time_at_start;
+    LD->profile.time += tend - LD->profile.time_at_start;
 
     stopItimer();
-    activateProfiler(PROF_INACTIVE PASS_LDARG(ld));
+    activateProfiler(PROF_INACTIVE);
 #ifndef __WINDOWS__
     set_sighandler(timer_signal, SIG_IGN);
     timer_signal = 0;
@@ -386,8 +407,9 @@ $prof_node(+Node, -Pred, -Calls, -Redos, -Exits,
 	   -Recursive, -Ticks, -SiblingTicks)
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define get_node(t, node) LDFUNC(get_node, t, node)
 static int
-get_node(term_t t, call_node **node ARG_LD)
+get_node(DECL_LD term_t t, call_node **node)
 { if ( PL_is_functor(t, FUNCTOR_dprof_node1) )
   { term_t a = PL_new_term_ref();
     void *ptr;
@@ -407,11 +429,12 @@ get_node(term_t t, call_node **node ARG_LD)
 }
 
 
+#define unify_node(t, node) LDFUNC(unify_node, t, node)
 static int
-unify_node(term_t t, call_node *node ARG_LD)
+unify_node(DECL_LD term_t t, call_node *node)
 { return PL_unify_term(t,
 		       PL_FUNCTOR, FUNCTOR_dprof_node1,
-		         PL_POINTER, node);
+			 PL_POINTER, node);
 }
 
 
@@ -426,15 +449,15 @@ PRED_IMPL("$prof_sibling_of", 2, prof_sibling_of, PL_FA_NONDETERMINISTIC)
     { atom_t a;
 
       if ( !PL_is_variable(A1) )
-      { if ( get_node(A1, &sibling PASS_LD) )
+      { if ( get_node(A1, &sibling) )
 	{ if ( sibling->parent )
-	    return unify_node(A2, sibling->parent PASS_LD);
+	    return unify_node(A2, sibling->parent);
 	}
 	fail;
       } else
       { if ( PL_get_atom(A2, &a) && a == ATOM_minus )
 	  sibling = LD->profile.roots;
-	else if ( get_node(A2, &parent PASS_LD) )
+	else if ( get_node(A2, &parent) )
 	  sibling = parent->siblings;
 	else
 	  fail;
@@ -449,7 +472,7 @@ PRED_IMPL("$prof_sibling_of", 2, prof_sibling_of, PL_FA_NONDETERMINISTIC)
     { sibling = CTX_PTR;
 
     return_sibling:
-      if ( !unify_node(A1, sibling PASS_LD) )
+      if ( !unify_node(A1, sibling) )
 	fail;
       if ( sibling->next )
 	ForeignRedoPtr(sibling->next);
@@ -485,18 +508,18 @@ PRED_IMPL("$prof_node", 8, prof_node, 0)
 { PRED_LD
   call_node *n = NULL;
 
-  if ( !get_node(A1, &n PASS_LD) )
+  if ( !get_node(A1, &n) )
     return FALSE;
 
-  collectSiblingsTime(PASS_LD1);
+  collectSiblingsTime();
 
   return ( unify_node_id(A2, n) &&
-	   PL_unify_integer(A3, n->calls) &&
-	   PL_unify_integer(A4, n->redos) &&
-	   PL_unify_integer(A5, n->exits) &&
-	   PL_unify_integer(A6, n->recur) &&
-	   PL_unify_integer(A7, n->ticks) &&
-	   PL_unify_integer(A8, n->sibling_ticks) );
+	   PL_unify_uint64(A3, n->calls) &&
+	   PL_unify_uint64(A4, n->exits + n->fails - n->calls) &&
+	   PL_unify_uint64(A5, n->exits) &&
+	   PL_unify_uint64(A6, n->recur) &&
+	   PL_unify_uint64(A7, n->ticks) &&
+	   PL_unify_uint64(A8, n->sibling_ticks) );
 }
 
 
@@ -518,22 +541,22 @@ typedef struct prof_ref
   void * handle;			/* Procedure handle */
   PL_prof_type_t   *type;
   int   cycle;
-  uintptr_t ticks;
-  uintptr_t sibling_ticks;
-  uintptr_t calls;			/* calls to/from this predicate */
-  uintptr_t redos;			/* redos to/from this predicate */
-  uintptr_t exits;			/* exits to/from this predicate */
+  uint64_t ticks;
+  uint64_t sibling_ticks;
+  uint64_t calls;			/* calls to/from this predicate */
+  uint64_t redos;			/* redos to/from this predicate */
+  uint64_t exits;			/* exits to/from this predicate */
 } prof_ref;
 
 
 typedef struct
 { Definition def;
-  uintptr_t ticks;
-  uintptr_t sibling_ticks;
-  uintptr_t calls;
-  uintptr_t redos;
-  uintptr_t exits;
-  uintptr_t recur;
+  uint64_t ticks;
+  uint64_t sibling_ticks;
+  uint64_t calls;
+  uint64_t redos;
+  uint64_t exits;
+  uint64_t recur;
   prof_ref *callers;
   prof_ref *callees;
 } node_sum;
@@ -562,13 +585,13 @@ add_parent_ref(node_sum *sum,
 { prof_ref *r;
 
   sum->calls += self->calls;
-  sum->redos += self->redos;
+  sum->redos += self->exits + self->fails - self->calls;
   sum->exits += self->exits;
 
   for(r=sum->callers; r; r=r->next)
   { if ( r->handle == handle && r->cycle == cycle )
     { r->calls += self->calls;
-      r->redos += self->redos;
+      r->redos += self->exits + self->fails - self->calls;
       r->exits += self->exits;
       r->ticks += self->ticks;
       r->sibling_ticks += self->sibling_ticks;
@@ -579,7 +602,7 @@ add_parent_ref(node_sum *sum,
 
   r = allocHeapOrHalt(sizeof(*r));
   r->calls = self->calls;
-  r->redos = self->redos;
+  r->redos = self->exits + self->fails - self->calls;
   r->exits = self->exits;
   r->ticks = self->ticks;
   r->sibling_ticks = self->sibling_ticks;
@@ -592,7 +615,7 @@ add_parent_ref(node_sum *sum,
 
 
 static void
-add_recursive_ref(node_sum *sum, uintptr_t count, int cycle)
+add_recursive_ref(node_sum *sum, uint64_t count, int cycle)
 { prof_ref *r;
 
   for(r=sum->callers; r; r=r->next)
@@ -620,7 +643,7 @@ add_sibling_ref(node_sum *sum, call_node *sibling, int cycle)
   for(r=sum->callees; r; r=r->next)
   { if ( r->handle == sibling->handle && r->cycle == cycle )
     { r->calls += sibling->calls;
-      r->redos += sibling->redos;
+      r->redos += sibling->exits + sibling->fails - sibling->calls;
       r->exits += sibling->exits;
       r->ticks += sibling->ticks;
       r->sibling_ticks += sibling->sibling_ticks;
@@ -631,7 +654,7 @@ add_sibling_ref(node_sum *sum, call_node *sibling, int cycle)
 
   r = allocHeapOrHalt(sizeof(*r));
   r->calls = sibling->calls;
-  r->redos = sibling->redos;
+  r->redos = sibling->exits + sibling->fails - sibling->calls;
   r->exits = sibling->exits;
   r->ticks = sibling->ticks;
   r->sibling_ticks = sibling->sibling_ticks;
@@ -645,9 +668,10 @@ add_sibling_ref(node_sum *sum, call_node *sibling, int cycle)
 
 
 
+#define sumProfile(n, handle, type, sum, seen) LDFUNC(sumProfile, n, handle, type, sum, seen)
 static int
-sumProfile(call_node *n, void *handle, PL_prof_type_t *type,
-	   node_sum *sum, int seen ARG_LD)
+sumProfile(DECL_LD call_node *n, void *handle, PL_prof_type_t *type,
+	   node_sum *sum, int seen)
 { call_node *s;
   int count = 0;
 
@@ -673,21 +697,18 @@ sumProfile(call_node *n, void *handle, PL_prof_type_t *type,
   }
 
   for(s=n->siblings; s; s = s->next)
-    count += sumProfile(s, handle, type, sum, seen PASS_LD);
+    count += sumProfile(s, handle, type, sum, seen);
 
   return count;
 }
 
 
+#define unify_relatives(list, r) LDFUNC(unify_relatives, list, r)
 static int
-unify_relatives(term_t list, prof_ref *r ARG_LD)
+unify_relatives(DECL_LD term_t list, prof_ref *r)
 { term_t tail = PL_copy_term_ref(list);
   term_t head = PL_new_term_ref();
   term_t tmp = PL_new_term_ref();
-  static functor_t FUNCTOR_node7;
-
-  if ( !FUNCTOR_node7 )
-    FUNCTOR_node7 = PL_new_functor(PL_new_atom("node"), 7);
 
   for( ; r; r=r->next)
   { int rc;
@@ -707,11 +728,11 @@ unify_relatives(term_t list, prof_ref *r ARG_LD)
 	 !PL_unify_term(head, PL_FUNCTOR, FUNCTOR_node7,
 			PL_TERM, tmp,
 			PL_INT,  r->cycle,
-			PL_LONG, r->ticks,
-			PL_LONG, r->sibling_ticks,
-			PL_LONG, r->calls,
-			PL_LONG, r->redos,
-		        PL_LONG, r->exits) )
+			PL_INT64, r->ticks,
+			PL_INT64, r->sibling_ticks,
+			PL_INT64, r->calls,
+			PL_INT64, r->redos,
+			PL_INT64, r->exits) )
       fail;
   }
 
@@ -752,6 +773,14 @@ get_handle(term_t t, void **handle)
   fail;
 }
 
+static void
+prof_release_def(void *handle)
+{ Definition def = handle;
+
+  releaseModule(def->module);
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 $prof_procedure_data(+PredicateIndicator,
 		     -Ticks, -TicksSiblings,
@@ -772,21 +801,21 @@ PRED_IMPL("$prof_procedure_data", 8, prof_procedure_data, PL_FA_TRANSPARENT)
   if ( !get_handle(A1, &handle) )
     fail;
 
-  collectSiblingsTime(PASS_LD1);
+  collectSiblingsTime();
   memset(&sum, 0, sizeof(sum));
   for(n=LD->profile.roots; n; n=n->next)
-    count += sumProfile(n, handle, &prof_default_type, &sum, 0 PASS_LD);
+    count += sumProfile(n, handle, &prof_default_type, &sum, 0);
 
   if ( count == 0 )
     fail;				/* nothing known about this one */
 
-  rc = ( PL_unify_integer(A2, sum.ticks) &&
-	 PL_unify_integer(A3, sum.sibling_ticks) &&
-	 PL_unify_integer(A4, sum.calls) &&
-	 PL_unify_integer(A5, sum.redos) &&
-	 PL_unify_integer(A6, sum.exits) &&
-	 unify_relatives(A7, sum.callers PASS_LD) &&
-	 unify_relatives(A8, sum.callees PASS_LD)
+  rc = ( PL_unify_uint64(A2, sum.ticks) &&
+	 PL_unify_uint64(A3, sum.sibling_ticks) &&
+	 PL_unify_uint64(A4, sum.calls) &&
+	 PL_unify_uint64(A5, sum.redos) &&
+	 PL_unify_uint64(A6, sum.exits) &&
+	 unify_relatives(A7, sum.callers) &&
+	 unify_relatives(A8, sum.callees)
        );
 
   free_relatives(sum.callers);
@@ -796,26 +825,33 @@ PRED_IMPL("$prof_procedure_data", 8, prof_procedure_data, PL_FA_TRANSPARENT)
 }
 
 
-/** '$prof_statistics'(-Samples, -Ticks, -AccountingTicks, -Time, -Nodes)
+/** '$prof_statistics'(-Samples, -Ticks, -AccountingTicks,
+		       -Time, -Nodes, -Period, -Ports)
 
 @arg Samples is the number of times the statistical profiler was called.
 @arg Ticks   is the number of virtual ticks during profiling
 @arg AccountingTicks are tick spent on accounting
 @arg Time    is the total CPU time spent profiling
 @arg Nodes   is the number of nodes in the call tree
+@arg Period  is the number of microseconds between samples
+@arg Ports   is one of `true`, `false` or `classic`
 */
 
 static
-PRED_IMPL("$prof_statistics", 5, prof_statistics, 0)
+PRED_IMPL("$prof_statistics", 7, prof_statistics, 0)
 { PRED_LD
-  if ( PL_unify_integer(A1, LD->profile.samples) &&
-       PL_unify_integer(A2, LD->profile.ticks) &&
-       PL_unify_integer(A3, LD->profile.accounting_ticks) &&
-       PL_unify_float(  A4, LD->profile.time) &&
-       PL_unify_integer(A5, LD->profile.nodes) )
-    succeed;
 
-  fail;
+  return
+    ( PL_unify_uint64(A1, LD->profile.samples) &&
+      PL_unify_uint64(A2, LD->profile.ticks) &&
+      PL_unify_uint64(A3, LD->profile.accounting_ticks) &&
+      PL_unify_float( A4, LD->profile.time) &&
+      PL_unify_uint64(A5, LD->profile.nodes) &&
+      PL_unify_integer(A6, LD->profile.sample_period) &&
+      PL_unify_atom(A7, LD->profile.ports_control == PROFC_FALSE ? ATOM_false :
+			LD->profile.ports_control == PROFC_TRUE  ? ATOM_true :
+								   ATOM_classic)
+    );
 }
 
 
@@ -823,8 +859,14 @@ PRED_IMPL("$prof_statistics", 5, prof_statistics, 0)
 		 *	       RESET		*
 		 *******************************/
 
+#if USE_LD_MACROS
+#define prof_clear_environments(fr) LDFUNC(prof_clear_environments, fr)
+#define prof_clear_choicepoints(ch) LDFUNC(prof_clear_choicepoints, ch)
+#define prof_clear_stacks(fr, ch) LDFUNC(prof_clear_stacks, fr, ch)
+#endif
+
 static QueryFrame
-prof_clear_environments(PL_local_data_t *ld, LocalFrame fr)
+prof_clear_environments(DECL_LD LocalFrame fr)
 { if ( fr == NULL )
     return NULL;
 
@@ -832,7 +874,7 @@ prof_clear_environments(PL_local_data_t *ld, LocalFrame fr)
   { if ( true(fr, FR_MARKED) )
       return NULL;
     set(fr, FR_MARKED);
-    ld->gc._local_frames++;
+    LD->gc._local_frames++;
 
     fr->prof_node = NULL;
 
@@ -845,22 +887,22 @@ prof_clear_environments(PL_local_data_t *ld, LocalFrame fr)
 
 
 static void
-prof_clear_choicepoints(PL_local_data_t *ld, Choice ch)
+prof_clear_choicepoints(DECL_LD Choice ch)
 { for( ; ch; ch = ch->parent )
-  { ld->gc._choice_count++;
-    prof_clear_environments(ld, ch->frame);
+  { LD->gc._choice_count++;
+    prof_clear_environments(ch->frame);
   }
 }
 
 
 static void
-prof_clear_stacks(PL_local_data_t *ld, LocalFrame fr, Choice ch)
+prof_clear_stacks(DECL_LD LocalFrame fr, Choice ch)
 { QueryFrame qf;
 
   while(fr)
-  { qf = prof_clear_environments(ld, fr);
+  { qf = prof_clear_environments(fr);
     assert(qf->magic == QID_MAGIC);
-    prof_clear_choicepoints(ld, ch);
+    prof_clear_choicepoints(ch);
     if ( qf->parent )
     { QueryFrame pqf = qf->parent;
 
@@ -874,15 +916,14 @@ prof_clear_stacks(PL_local_data_t *ld, LocalFrame fr, Choice ch)
 
 
 bool
-resetProfiler(void)
-{ GET_LD
-  stopProfiler();
+resetProfiler(DECL_LD)
+{ stopProfiler();
 
   assert(LD->gc._local_frames == 0);
   assert(LD->gc._choice_count == 0);
 
-  prof_clear_stacks(LD, environment_frame, LD->choicepoints);
-  unmark_stacks(LD, environment_frame, LD->choicepoints, FR_MARKED);
+  prof_clear_stacks(environment_frame, LD->choicepoints);
+  unmark_stacks(environment_frame, LD->choicepoints, FR_MARKED);
 
   assert(LD->gc._local_frames == 0);
   assert(LD->gc._choice_count == 0);
@@ -911,13 +952,28 @@ PRED_IMPL("reset_profiler", 0, reset_profiler, 0)
 		 *******************************/
 
 static
-PRED_IMPL("$profile", 2, profile, PL_FA_TRANSPARENT)
+PRED_IMPL("$profile", 4, profile, PL_FA_TRANSPARENT)
 { int rc;
   prof_status val;
 
   if ( !get_prof_status(A2, &val) )
     return FALSE;
 
+  atom_t ports_opt;
+  PL_get_atom_ex(A3, &ports_opt);
+  switch (ports_opt)
+  { case ATOM_false: LD->profile.ports_control = PROFC_FALSE;   break;
+    case ATOM_true:  LD->profile.ports_control = PROFC_TRUE;    break;
+    default:         LD->profile.ports_control = PROFC_CLASSIC;
+  }
+
+  LD->profile.sample_period = 5000;  // default period in usecs.
+  double rate;
+  if ( PL_get_float_ex(A4, &rate) )
+    LD->profile.sample_period = (unsigned int)(1000000/rate);
+  else
+    return FALSE;
+
   resetProfiler();
   startProfiler(val);
   rc = callProlog(NULL, A1, PL_Q_PASS_EXCEPTION, NULL);
@@ -941,7 +997,7 @@ statistics.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static void
-profile(intptr_t count, PL_local_data_t *__PL_ld)
+profile(DECL_LD intptr_t count)
 { call_node *node;
 
   if ( !HAS_LD )
@@ -995,8 +1051,9 @@ node_name(call_node *n)
 }
 #endif
 
+#define prof_call(handle, type) LDFUNC(prof_call, handle, type)
 static call_node *
-prof_call(void *handle, PL_prof_type_t *type ARG_LD)
+prof_call(DECL_LD void *handle, PL_prof_type_t *type)
 { call_node *node = LD->profile.current;
 
   LD->profile.accounting = TRUE;
@@ -1039,7 +1096,7 @@ prof_call(void *handle, PL_prof_type_t *type ARG_LD)
 	  Sdprintf("Call: direct recursion on %s\n", node_name(node)));
     LD->profile.accounting = FALSE;
     return node;
-  } else				/* from some parent */
+  } else if ( LD->profile.ports_control != PROFC_TRUE )
   { void *parent = node->handle;
 
     for(node=node->parent; node; node = node->parent)
@@ -1088,11 +1145,13 @@ prof_call(void *handle, PL_prof_type_t *type ARG_LD)
 
 
 call_node *
-profCall(Definition def ARG_LD)
+profCall(DECL_LD Definition def)
 { if ( true(def, P_NOPROFILE) )
     return LD->profile.current;
 
-  return prof_call(def, &prof_default_type PASS_LD);
+  acquireModulePtr(def->module);
+
+  return prof_call(def, &prof_default_type);
 }
 
 
@@ -1107,60 +1166,93 @@ code.  Considering this is development only, we'll leave this for now.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 void
-profResumeParent(struct call_node *node ARG_LD)
+profResumeParent(DECL_LD struct call_node *node)
 { call_node *n;
 
   if ( node && node->magic != PROFNODE_MAGIC )
     return;
 
   LD->profile.accounting = TRUE;
-  for(n=LD->profile.current; n && n != node; n=n->parent)
-  { DEBUG(MSG_PROF_CALLTREE,
-	  Sdprintf("Exit: %s\n", node_name(n)));
-    n->exits++;
-  }
-  LD->profile.accounting = FALSE;
 
-  LD->profile.current = node;
-}
-
-
-void
-profExit(struct call_node *node ARG_LD)
-{ if ( !node || node->magic != PROFNODE_MAGIC )
-    return;
-
-  profResumeParent(node->parent PASS_LD);
-}
-
-
-void
-profRedo(struct call_node *node ARG_LD)
-{ if ( node && node->magic != PROFNODE_MAGIC )
-    return;
-
-  if ( node )
-  { if ( LD->profile.current )
-    { struct call_node *n;
-
-      DEBUG(MSG_PROF_CALLTREE,
-	    Sdprintf("Redo: on %s; current is %s\n",
-		     node_name(node), node_name(LD->profile.current) ));
-
-      for(n=node; n && n != LD->profile.current; n = n->parent)
-      { DEBUG(MSG_PROF_CALLTREE,
-	      Sdprintf("Redo: %s\n", node_name(n)));
-	n->redos++;
-      }
-    } else
+  if ( LD->profile.ports_control  != PROFC_FALSE )
+  { for(n=LD->profile.current; n && n != node; n=n->parent)
     { DEBUG(MSG_PROF_CALLTREE,
-	    Sdprintf("Redo: %s\n", node_name(node)));
-      node->redos++;
+	    Sdprintf("Exit: %s\n", node_name(n)));
+      n->exits++;
     }
   }
   LD->profile.current = node;
+
+  LD->profile.accounting = FALSE;
+
 }
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+
+profExit(call_node from_node)
+
+Exit to the parent of from_node. Note that all nodes from the current
+node back to and including from_node are deemed to have exited. This
+catches all normal exits (from_node == current) and last calls (from_node
+is an ancestor of current).
+
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+void
+profExit(DECL_LD struct call_node *node)
+{ if ( !node || node->magic != PROFNODE_MAGIC )
+    return;
+
+  profResumeParent(node->parent);
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+profFail(call_node cp_node)
+
+Fail to choicepoint node. All nodes in  the branch from the current node
+back to a common ancestor with cp_node   are deemed to have failed. Note
+that if the current node is the  choicepoint node, nothing fails as this
+is internal to the predicate (doesn't pass through Fail port).
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static int
+lookback(call_node *from_node, call_node *to_node)
+{ struct call_node *node = from_node;
+
+  while( node )
+  { if ( node == to_node )
+      return TRUE;
+    else
+      node = node->parent;
+  }
+
+  return FALSE;
+}
+
+static void
+profFailToCP(struct call_node *current, struct call_node *cp_node)
+{ while (current && (!lookback(cp_node, current)))
+  { DEBUG(MSG_PROF_CALLTREE,
+	  Sdprintf("Fail: %s\n", node_name(current)));
+
+    if (current->magic == PROFNODE_MAGIC)
+      current->fails++;
+
+    current = current->parent;
+  }
+}
+
+void
+profFail(DECL_LD struct call_node *cp_node)
+{ LD->profile.accounting = TRUE;
+
+  if ( LD->profile.ports_control != PROFC_FALSE )
+    profFailToCP(LD->profile.current, cp_node);
+  LD->profile.current = cp_node;
+
+  LD->profile.accounting = FALSE;
+}
 
 void
 profSetHandle(struct call_node *node, void *handle)
@@ -1197,7 +1289,7 @@ void *
 PL_prof_call(void *handle, PL_prof_type_t *type)
 { GET_LD
 
-  return prof_call(handle, type PASS_LD);
+  return prof_call(handle, type);
 }
 
 void
@@ -1205,7 +1297,7 @@ PL_prof_exit(void *node)
 { GET_LD
   struct call_node *n = node;
 
-  profResumeParent(n->parent PASS_LD);
+  profResumeParent(n->parent);
 }
 
 
@@ -1214,10 +1306,10 @@ PL_prof_exit(void *node)
 		 *******************************/
 
 
-static uintptr_t
+static uint64_t
 collectSiblingsNode(call_node *n)
 { call_node *s;
-  uintptr_t count = 0;
+  uint64_t count = 0;
 
   for(s=n->siblings; s; s=s->next)
   { count += collectSiblingsNode(s);
@@ -1229,7 +1321,7 @@ collectSiblingsNode(call_node *n)
 
 
 static void
-collectSiblingsTime(ARG1_LD)
+collectSiblingsTime(DECL_LD)
 { if ( !LD->profile.sum_ok )
   { call_node *n;
 
@@ -1241,8 +1333,9 @@ collectSiblingsTime(ARG1_LD)
 }
 
 
+#define freeProfileNode(node) LDFUNC(freeProfileNode, node)
 static void
-freeProfileNode(call_node *node ARG_LD)
+freeProfileNode(DECL_LD call_node *node)
 { call_node *n, *next;
 
   assert(node->magic == PROFNODE_MAGIC);
@@ -1250,7 +1343,10 @@ freeProfileNode(call_node *node ARG_LD)
   for(n=node->siblings; n; n=next)
   { next = n->next;
 
-    freeProfileNode(n PASS_LD);
+    if ( n->type && n->type->release )
+      (*n->type->release)(n->handle);
+
+    freeProfileNode(n);
   }
 
   node->magic = 0;
@@ -1270,64 +1366,17 @@ freeProfileData(void)
 
   for(; n; n=next)
   { next = n->next;
-    freeProfileNode(n PASS_LD);
+    freeProfileNode(n);
   }
 
   assert(LD->profile.nodes == 0);
 }
 
-#else /* O_PROFILE */
-
-		 /*******************************
-		 *	    NO PROFILER		*
-		 *******************************/
-
-void
-stopItimer(void)
-{
-}
-
-static
-PRED_IMPL("profiler", 2, profiler, 0)
-{ return notImplemented("profile", 2);
-}
-
-static
-PRED_IMPL("reset_profiler", 0, reset_profiler, 0)
-{ return notImplemented("reset_profile", 0);
-}
-
-static
-PRED_IMPL("$prof_node", 8, prof_node, 0)
-{ return notImplemented("profile_node", 8);
-}
-
-static
-PRED_IMPL("$prof_sibling_of", 2, prof_sibling_of, PL_FA_NONDETERMINISTIC)
-{ return notImplemented("profile_sibling_of", 2);
-}
-
-static
-PRED_IMPL("$profile", 2, profile, PL_FA_TRANSPARENT)
-{ return notImplemented("$profile", 2);
-}
-
-static
-PRED_IMPL("$prof_procedure_data", 8, prof_procedure_data, PL_FA_TRANSPARENT)
-{ return notImplemented("$prof_procedure_data", 8);
-}
-
-static
-PRED_IMPL("$prof_statistics", 5, prof_statistics, 0)
-{ return notImplemented("$prof_statistics", 5);
-}
-
-/* Foreign interface of the profiler
-*/
+#else /* !O_PROFILE */
 
 int
 PL_register_profile_type(PL_prof_type_t *type)
-{ return FALSE;				/* not supported */
+{ return FALSE;
 }
 
 void *
@@ -1340,38 +1389,21 @@ PL_prof_exit(void *node)
 {
 }
 
+
 #endif /* O_PROFILE */
 
-#ifdef O_PROF_PENTIUM
-#include "pentium.c"
-
-PRED_IMPL("show_pentium_profile", 0, show_pentium_profile, 0)
-{ prof_report();
-
-  succeed;
-}
-
-PRED_IMPL("reset_pentium_profile", 0, reset_pentium_profile, 0)
-{ prof_reset();
-
-  succeed;
-}
-#endif
-
 		 /*******************************
 		 *      PUBLISH PREDICATES	*
 		 *******************************/
 
 BeginPredDefs(profile)
-  PRED_DEF("$profile", 2, profile, PL_FA_TRANSPARENT)
+#if O_PROFILE
+  PRED_DEF("$profile", 4, profile, PL_FA_TRANSPARENT)
   PRED_DEF("profiler", 2, profiler, 0)
   PRED_DEF("reset_profiler", 0, reset_profiler, 0)
   PRED_DEF("$prof_node", 8, prof_node, 0)
   PRED_DEF("$prof_sibling_of", 2, prof_sibling_of, PL_FA_NONDETERMINISTIC)
   PRED_DEF("$prof_procedure_data", 8, prof_procedure_data, PL_FA_TRANSPARENT)
-  PRED_DEF("$prof_statistics", 5, prof_statistics, 0)
-#ifdef O_PROF_PENTIUM
-  PRED_DEF("show_pentium_profile", 0, show_pentium_profile, 0)
-  PRED_DEF("reset_pentium_profile", 0, reset_pentium_profile, 0)
+  PRED_DEF("$prof_statistics", 7, prof_statistics, 0)
 #endif
 EndPredDefs
